from crostab import Crostab

from pyspare.margin.matrix_margin import matrix_margin
from pyspare.margin.vector_margin import vector_margin


def crostab_margin(crostab: Crostab, top, bottom, left, right, height, width, read, side_read, head_read, rule):
    # print(matrix_margin(crostab.rows, top, bottom, left, right, height, width, read, rule))
    return Crostab(
        side=vector_margin(crostab.side, top, bottom, side_read, rule),
        head=vector_margin(crostab.head, left, right, head_read, rule),
        rows=matrix_margin(crostab.rows, top, bottom, left, right, height, width, read, rule),
        title=crostab.title
    )



from veho.matrix import size


def sizing(matrix, top, bottom, left, right, height=None, width=None):
    if not height or not width: (height, width) = size(matrix)
    if not height or not width: top, bottom = 0, 0
    if (not top and not bottom) or (top + bottom >= height): top, bottom = height, 0
    if (not left and not right) or (left + right >= width): left, right = width, 0
    return top, bottom, left, right, height, width



from aryth.bound_vector import max_by
from crostab import Series
from texting import DA, lange, to_lpad, to_rpad
from veho.vector import mapper


def series_padder(series: Series, ansi, full_angle):
    lpad = to_lpad(ansi=ansi)
    rpad = to_rpad(ansi=ansi)
    length = lange if ansi else len
    pad = max(length(series.title), max_by(series.points, length))
    series = Series(
        title=rpad(series.title, pad),
        rule=DA * pad,
        points=mapper(series.points, lambda x: lpad(x, pad))
    )
    # print(series)
    return series



from crostab import Table

from pyspare.margin.matrix_margin import matrix_margin
from pyspare.margin.vector_margin import vector_margin


def table_margin(table: Table, top, bottom, left, right, height, width, read, head_read, rule):
    return Table(
        head=vector_margin(table.head, left, right, head_read, rule),
        rows=matrix_margin(table.rows, top, bottom, left, right, height, width, read, rule),
        title=table.title
    )



from ject import oneself
from palett import fluo_matrix, fluo_vector
from palett.presets import FRESH, PLANET
from texting import AEU, COSP, ELLIP, liner
from texting.enum.brackets import BRK
from veho.enum.matrix_directions import COLUMNWISE
from veho.matrix import size
from veho.vector import mapper

from pyspare.margin import MatrixMargin, VectorMargin, table_margin
from pyspare.padder import table_padder
from pyspare.padder.matrix_padder import matrix_padder


def deco_table(table,
               discrete=False,
               read=None,
               head_read=None,
               presets=(FRESH, PLANET),
               direct=COLUMNWISE,
               top=0,
               bottom=0,
               left=0,
               right=0,
               ansi=False,
               full_angle=False,
               level=0,
               rule=ELLIP):
    height, width = size(table.rows)
    label_width = len(table.head)
    if not height or not width or not label_width: return AEU
    table = table_margin(table, top, bottom, left, right, height, width, read, head_read, rule)
    table = table_padder(table, ansi, full_angle)
    if presets:
        table = table.boot(
            head=fluo_vector(table.head, presets, mutate=True),
            rows=fluo_matrix(table.rows, direct, presets, mutate=True)
        )
    lines = [' | '.join(table.head),
             '-+-'.join(table.rule)]
    lines.extend(mapper(table.rows,
                        lambda row: ' | '.join(row)))
    return liner(lines, level=level, discrete=discrete)


def deco_table_arch(table,
                    discrete=False,
                    delim=COSP,
                    bracket=BRK,
                    read=None,
                    head_read=None,
                    presets=(FRESH, PLANET),
                    direct=COLUMNWISE,
                    top=0,
                    bottom=0,
                    left=0,
                    right=0,
                    ansi=False,
                    full_angle=False,
                    level=0,
                    hr=ELLIP):
    head = table['head'], rows = table['rows']
    height, width = size(rows)
    label_width = len(head)
    if not height or not width or not label_width: return AEU
    head_margin, rows_margin = (VectorMargin.build(head, left, right),
                                MatrixMargin.build(rows, top, bottom, left, right, height, width))
    head_raw, rows_raw = (head_margin.map(oneself).to_list(ELLIP),
                          rows_margin.map(oneself).to_matrix(hr))
    head_alt, rows_alt = (head_margin.stringify(read).to_list(ELLIP),
                          rows_margin.stringify(read).to_matrix(hr))
    rows_dye, head_dye = (fluo_vector(head_raw, direct, presets, colorant=True),
                          fluo_matrix(rows_raw, direct, presets, colorant=True)) \
        if presets else (None, None)
    rows = matrix_padder(rows_alt, rows_raw, rows_dye, ansi)



from aryth.bound_entries import max_by
from texting import SP, lange, to_lpad, to_pad
from veho.entries import mapper


def entries_padder(entries, ansi=False, fill=SP):
    lpad, pad = to_lpad(ansi=ansi, fill=fill), to_pad(ansi=ansi, fill=fill)
    kw, vw = max_by(entries, lange if ansi else len)
    return mapper(entries, lambda x: lpad(x, kw), lambda tx: pad(tx, vw, tx))



from .crostab_padder import crostab_padder
from .entries_padder import entries_padder
from .matrix_padder import matrix_padder
from .series_padder import series_padder
from .table_padder import table_padder
from .vector_padder import vector_padder


def pad_start(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}>{width}}'


def pad_end(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}<{width}}'


def pad_centered(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}^{width}}'



from .crostab_margin import crostab_margin
from .entries_margin import EntriesMargin, entries_margin
from .matrix_margin import MatrixMargin, matrix_margin
from .table_margin import table_margin
from .vector_margin import VectorMargin, vector_margin



from crostab import Crostab
from palett import fluo_matrix, fluo_vector
from palett.presets import FRESH, PLANET
from texting import AEU, ELLIP, liner
from veho.enum.matrix_directions import COLUMNWISE
from veho.matrix import size
from veho.vector import zipper

from pyspare import deco_matrix, deco_vector
from pyspare.margin.crostab_margin import crostab_margin
from pyspare.padder.crostab_padder import crostab_padder


def deco_crostab(crostab: Crostab,
                 discrete=False,
                 read=None,
                 side_read=None,
                 head_read=None,
                 presets=(FRESH, PLANET),
                 direct=COLUMNWISE,
                 top=0,
                 bottom=0,
                 left=0,
                 right=0,
                 ansi=False,
                 full_angle=False,
                 level=0,
                 rule=ELLIP):
    height, width = size(crostab.rows)
    label_height, label_width = len(crostab.side), len(crostab.head)
    if not height or not width or not label_width or not label_height: return AEU
    crostab = crostab_margin(crostab, top, bottom, left, right, height, width, read, side_read, head_read, rule)
    # print(deco_vector(crostab.head))
    # print(deco_vector(crostab.side))
    # print(deco_matrix(crostab.rows))
    crostab = crostab_padder(crostab, ansi, full_angle)
    if presets:
        # vector_presets = (presets[0], presets[1])
        crostab = crostab.boot(
            side=fluo_vector(crostab.side, presets, mutate=True),
            head=fluo_vector(crostab.head, presets, mutate=True),
            rows=fluo_matrix(crostab.rows, direct, presets, mutate=True)
        )
    lines = [crostab.title + ' | ' + ' | '.join(crostab.head),
             '-+-'.join(crostab.rule)]
    lines.extend(zipper(crostab.side, crostab.rows,
                        lambda s, r: s + ' | ' + ' | '.join(r)))
    return liner(lines, level=level, discrete=discrete)



from aryth.bound_vector import max_by
from texting import SP, lange, to_pad
from veho.columns import mapper as mapper_columns
from veho.matrix.enumerate import mapper


def matrix_padder(matrix, ansi=False, fill=SP):
    pad = to_pad(ansi=ansi, fill=fill)
    length = lange if ansi else len
    widths = mapper_columns(matrix, lambda col: max_by(col, indicator=length))
    return mapper(matrix, lambda x, i, j: pad(x, widths[j], x))



from .sizing import sizing
from .vector_margin import VectorMargin


def vector_margin(vec, head, tail, read, rule):
    return VectorMargin \
        .build(vec, head, tail) \
        .stringify(read) \
        .to_list(rule)



from typing import Callable, List, Tuple

from ject import oneself
from palett import Preset, fluo_vector
from palett.presets import FRESH, PLANET
from texting import ELLIP, liner
from texting.enum.brackets import BRK
from veho.vector import mutazip
from veho.vector.length import length

from pyspare.margin import VectorMargin, vector_margin


def deco_vector(
        vec: list,
        read: Callable = None,
        head: int = None,
        tail: int = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim=',\n',
        bracket=BRK,
        discrete=False
):
    size = length(vec)
    if not size: return str(vec)
    vec = vector_margin(vec, head, tail, read, ELLIP)
    # vn = VectorMargin.build(vec, head, tail)
    # raw, matrix = vn.map(oneself).to_list(ELLIP), vn.stringify(read).to_list(ELLIP)
    if presets:
        vec = fluo_vector(vec, presets, effects, colorant=False, mutate=True)
        # matrix = mutazip(matrix, dyes, lambda tx, dy: dy(tx))
    return liner(vec, delim=delim, bracket=bracket, discrete=discrete)


def deco_vector_arch(
        vec: list,
        read: Callable = None,
        head: int = None,
        tail: int = None,
        presets: Tuple[Preset] = (FRESH, PLANET),
        effects: List[str] = None,
        delim=',\n',
        bracket=BRK,
        discrete=False
):
    size = length(vec)
    if not size: return str(vec)
    vn = VectorMargin.build(vec, head, tail)
    raw, text = vn.map(oneself).to_list(ELLIP), vn.stringify(read).to_list(ELLIP)
    if presets:
        dyes = fluo_vector(raw, presets, effects, colorant=True, mutate=True)
        text = mutazip(text, dyes, lambda tx, dy: dy(tx))
    return liner(text, delim=delim, bracket=bracket, discrete=discrete)



from .matrix_margin import MatrixMargin
from .sizing import sizing


def matrix_margin(mx, top, bottom, left, right, height, width, read, rule):
    return MatrixMargin \
        .build(mx, top, bottom, left, right, height, width) \
        .stringify(read) \
        .to_matrix(rule)



from crostab import Crostab, Series
from texting import SP

from pyspare.padder.series_padder import series_padder
from pyspare.padder.table_padder import table_padder


def crostab_padder(crostab: Crostab, ansi=False, full_angle=False, fill=SP):
    series = Series(points=crostab.side, title=crostab.title)
    side_part = series_padder(series, ansi, full_angle)
    body_part = table_padder(crostab, ansi, full_angle)
    return Crostab(
        side=side_part.points,
        head=body_part.head,
        rows=body_part.rows,
        title=side_part.title,
        rule=[side_part.rule] + body_part.rule
    )



from aryth.bound_vector import max_by
from crostab import Table
from texting import DA, SP, lange, to_pad
from veho.columns import mapper as mapper_columns
from veho.matrix.enumerate import mapper as mapper_matrix
from veho.vector import mapper, zipper


def table_padder(table: Table, ansi=False, full_angle=False, fill=SP):
    # if (full_angle):
    padder = to_pad(ansi=ansi, fill=fill)
    length = lange if ansi else len
    widths = mapper_columns([table.head] + table.rows, lambda col: max_by(col, indicator=length))
    return Table(
        head=zipper(table.head, widths, lambda x, p: padder(x, p, x)),
        rule=mapper(widths, lambda p: DA * p),
        rows=mapper_matrix(table.rows, lambda x, i, j: padder(x, widths[j], x)),
        title=table.title
    )



from veho.vector import length


def sizing(vec, head, tail):
    size = length(vec)
    if not size: head, tail = 0, 0
    elif (not head and not tail) or (head + tail >= size): head, tail = size, 0
    return head, tail



from pyspare.margin.entries_margin.entries_margin import EntriesMargin


def entries_margin(entries, head, tail, key_read, read, rule):
    return EntriesMargin \
        .build(entries, head, tail) \
        .stringify(key_read, read) \
        .to_list(rule)  # ('..', '..') if rule is None else rule
